﻿using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using System.Windows.Forms;
using System.Drawing;
using ProcessRubyScript;
using System.Text;

namespace CodeInspector
{
    public partial class Form1 : Form
    {
        /*
         * EDITS:
         * 
         * Altered GetData( ) to look for 'calculations' as a flag to stop grabbing the document boxes
         * in 'Field Data'
         * 
         * 06/21/17 Version 2.0:
         * 
         *      1) Added code to inspect for usage of eval().
         *      2) Look for method signatures with duplicate names in same document.
         *      3) Look for method signatures that need renaming/replacing using the MethodMasterFile.rbw.
         *      4) Use of name_block().
         *      5) Altered the check box control update check - alerts user to get the latest version.
         *      
         * 08/14/17:
         * 
         *  Commenting out 3), not sure how we are going to handle the various check_digit versions
         *  and the old amap vs new amap methods.
         *  
         *  Altered the eval Warning message to alert user to run the black box converter to check for
         *  conversion errors as I have not seen eval usage causing conversion trouble.
        */

        // For now, we are defaulting to 'true'.
        bool no_comments = true;

        bool fromEditor = false;
        string fname = null;
        string cur_dir = @"C:\temp\work";
        Data doc_info;

        // This is set by ProcessMethods( ).
        List<string> methList;

        // Actual location, provided we use this:
        //string methMasterFile = @"\\PDC\FormsProgramming\Shared Code\Calc Tools\MethodMasterFile.rbw";

        // DCS testing at work:
        //string methMasterFile = @"\\PDC\FormsProgramming\Shared Code\Calc Tool Test\MethodMasterFile.rbw";

        // My testing at work:
        //string methMasterFile = @"E:\CheckDigitVariations\MethodMasterFile.rbw";

        // Testing at home:
        //string methMasterFile = @"H:\CheckDigitVariations\MethodMasterFile.rbw";

        /*
         * This holds the data which will be used by ProcessData() in reporting:
         *
         * A) Boxes referenced in Ruby code that are not on the document. Class members used are:
         *
         * doc_box_list - A list of all boxes in a document.
         * code_box_list - A list of all boxes in a document, no duplicates.
         * code_box_lines - A dictionary where the key is a box from code_box_list with an
         *                  associated list of the document's Ruby code line numbers per key.
         *
         * B) Unbalanced print statements - i.e. print 'box1_1', box1_2.  Class members used are:
         *
         * unbal_print - A list of unbalanced box print statements with the associated Ruby code
         *                  line number.
         *
         * C) Check Box Control code needs updating. Class members used are:
         *
         * chk_box_ctrl_flag - Indicates that current check box control code needs to be updated to latest version.
         *
         * D) End of Data Flow comment is missing. Class members used are:
         *
         * no_eodf_comment - A boolean value, 'true' if end of data flow comment is missing else 'false'.
         *
         * E) Usage of eval() - Class members used:
         * 
         * evalList - Holds line numbers where eval() has been detected.
         * 
         * F) Use of name_block() method - Class members used:
         * 
         * methNames - We will inspect this for 'name_block', an old dataflow method.
         * 
         * G) Method name [the method itself not the invocation] appears in Ruby script more than once - Class members:
         * 
         * methNames - A dictionary that holds the method name [key] and a list of line number(s).
         * dupMethName - Boolean to indicate duplicate names detected in methNames.
         * 
         * H) Document methods that need to be updated to the Method Master File version. Note that this may
         *      only be a name change or a complete method replacement.
         * 
         * methUpdate - A string list that holds the document's mmethod name, line number of the 'def' statement,
         *              master file method name, and name of the master file category where the method master resides.
         *
         * I) Document methods that need to be added to the Method Master File version.
         * 
         * methAddMstr - A string list that holds the document's mmethod name, line number of the 'def' statement,
         *              and name of the master file category where the method master resides.
         *
         ****************************************************************** 
         * This Data class is written to by: GetData().
        */
        public class Data
        {
            public List<string> doc_box_list = new List<string>();
            public List<string> code_box_list = new List<string>();
            public List<string> unbal_print = new List<string>();
            public List<int> evalList = new List<int>();
            public Dictionary<string, List<int>> code_box_lines = new Dictionary<string, List<int>>();
            public List<string> methUpdate = new List<string>();
            public List<string> methAddMstr = new List<string>();
            public Dictionary<string, List<string>> methNames = new Dictionary<string, List<string>>();
            public bool dupMethName = false;
            public bool no_eodf_comment = false;
            public bool chk_box_ctrl_flag = false;
        }

        public void DisplayErrorMsg(string msg, string type)
        {
            outText.SelectionColor = (type == "ERR" ? Color.White : Color.White);
            outText.SelectionBackColor = Color.Black;
            outText.AppendText(msg);
            outText.SelectionColor = Color.Black;
            outText.SelectionBackColor = Color.White;
        }

        // This is called from Form1.cs [Form1()] only when command line args have been supplied, which is
        // currently expected to be just from the SciTE editor.
        //
        // When started from SciTE, there will be 2 command line arguments: the first one is the path
        // and file name of the document's Ruby code file being edited, the second argument is the string
        // "SciTE" which verifys that this application was started from SciTE, in which case the window title
        // reflects this fact to alert user that they need to end this app in order to use SciTE normally,
        // i.e. use <F5> or other Tools Menu items.
        //
        // 1 command line argument is processed when started from SciTE as a fall-thru in the below
        // switch statement. App could conceivably be started with just 1 command line argument which would be
        // the complete path and file name of a document's .RBW code file, though as indicated this is not
        // the intended usage currently. Intended usage is either a stand-alone app or started from SciTE.
        //
        // Also, the .RBW code file must be a document's Ruby code file, we just check the format of the
        // .RBW name - looking for 4 alpha chars followed by four digits with the '.rbw' extension.
        public string ProcessArgs(string[] vars)
        {
            // Holds just the file name and extension, not the path info.
            string tmp_file_name = null;

            string ret_fname = null;
            string err_msg = null;

            switch (vars.Length)
            {
                case (2):
                    //if ((vars[1].ToLower() == "scite") || (vars[1].ToLower() == "vsc"))
                    //{
                        // Using goto statement to fall through to next case label.
                        //fromEditor = true;
                    //}
                    //else
                    //{
                        //err_msg = "ERROR! Second command line argument to Code Inspector was neither \"SciTE\" nor \"VSC\".";
                        //break;
                    //}

                    if (vars[1].ToLower() != "scite")
                    {
                        err_msg = "ERROR! Second command line argument to Code Inspector was not \"SciTE\".";
                        break;
                    }
                    goto case (1);

                case (1):
                    try
                    {
                        // Added after removing 'VSC' from External Tools Extension:
                        fromEditor = true;

                        tmp_file_name = Path.GetFileName(vars[0]);
                    }
                    catch (ArgumentException argEx)
                    {
                        MessageBox.Show(argEx.ToString(), "Path parameter has invalid characters/is empty/contains only white spaces.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }

                    if (String.IsNullOrEmpty(tmp_file_name))
                    {
                        err_msg = "ERROR! Could not retrieve Ruby code file name from command line argument.";
                    }
                    else if (!Regex.IsMatch(tmp_file_name.ToLower(), @"^[a-z]{4}\d{4}\.(rb|rbw)$"))
                    {
                        err_msg = "ERROR! Code Inspector only inspects Ruby code files from a document. File name format does not follow document name format: Either 'XXXX9999.RBW' or 'XXXX9999.RB', where 'X' is any lower or upper case alpha character and '9' is any digit. The file extension may be upper or lower case.";
                    }
                    break;

                default:
                    err_msg = "ERROR! Code Inspector was started with the incorrect number of command line arguments.";
                    break;
            }

            fileName.Text = "Inspecting: " + (fromEditor ? vars[0] : string.Join(" ", vars));

            // Set the window title. Indicate if this tool was started from SciTE. Users need to
            // exit this application in order to use SciTE's Tool menu.
            this.Text = "Code Inspector 2.1" + (fromEditor ? " - Started From A Code Editor" : "");

            if (String.IsNullOrEmpty(err_msg))
            {
                ret_fname = vars[0];

                try
                {
                    cur_dir = Path.GetDirectoryName(vars[0]);
                }
                catch (ArgumentException ex)
                {
                    MessageBox.Show(ex.ToString(), "Path parameter has invalid characters/is empty/contains only white spaces.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (PathTooLongException ex)
                {
                    MessageBox.Show(ex.ToString(), "The path parameter is longer than the system-defined maximum length.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
            else
            {
                //These are all "ERR" type error messages, so we can do it like this else we would
                //need a var for the error type.
                DisplayErrorMsg(err_msg, "ERR");
            }

            return ret_fname;
        }

        // Called from Form1.cs [Form1()] openToolStripMenuItem_Click open file event handler. Returns
        // the full path and file name of the user selected Ruby file that belongs to a document.
        public string GetFileName()
        {
            string file_name = null;
            string err_msg;
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.InitialDirectory = cur_dir;
            ofd.Filter = "Ruby|*.rb|Ruby|*.rbw";

            System.Windows.Forms.DialogResult dr = ofd.ShowDialog();

            if (dr == DialogResult.OK)
            {
                // This holds the path and the file name with extension.
                file_name = ofd.FileName;

                fileName.Text = "Inspecting: " + ofd.FileName;
                outText.Clear();

                try
                {
                    if (!Regex.IsMatch(Path.GetFileName(file_name).ToLower(), @"^[a-z]{4}\d{4}\.(rb|rbw)$"))
                    {
                        err_msg = "ERROR! Code Inspector only inspects Ruby code files from a document. File name format does not follow document name format: Either 'XXXX9999.RBW' or 'XXXX9999.RB', where 'X' is any lower or upper case alpha character and '9' is any digit. The file extension may be upper or lower case.";
                        DisplayErrorMsg(err_msg, "ERR");
                        file_name = null;
                    }
                }
                catch (ArgumentException argEx)
                {
                    MessageBox.Show(argEx.ToString(), "Path parameter has invalid characters/is empty/contains only white spaces.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

                try
                {
                    cur_dir = Path.GetDirectoryName(ofd.FileName);
                }
                catch (ArgumentException ex)
                {
                    MessageBox.Show(ex.ToString(), "Path parameter has invalid characters/is empty/contains only white spaces.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                catch (PathTooLongException ex)
                {
                    MessageBox.Show(ex.ToString(), "The path parameter is longer than the system-defined maximum length.", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }

            return file_name;
        }

        // This is called from Form1.cs [Form1()] when command line args have been supplied.
        // Also called from Form1.cs [openToolStripMenuItem_Click()] after user selects a file
        // to inspect.
        public Data GetData(string fname, bool no_comments)
        {
            string flag = null;
            string line;
            //string methBody;
            string[] methParts;

            //ulong md5Hash;

            List<int> line_list = new List<int>();
            int ln_num = 0;

            bool profile_inst = false;
            bool end_of_df = false;
            bool chk_list = false;
            bool chk_list_join = false;

            Data doc_data = new Data();


            ProcessMethods(ref doc_data);

            try
            {
                // The using statement also closes the StreamReader.
                using (StreamReader sr = new StreamReader(fname))
                {
                    while ((line = sr.ReadLine()) != null)
                    {
                        ln_num++;

                        // Stripping out whitespace chars and converting to lowercase.
                        line = Regex.Replace(line.ToLower(), "[\n\r\t ]", string.Empty);

                        // Time to start grabbing document boxes?
                        if (Regex.IsMatch(line, @"^#\*+fielddata\*+$"))
                        {
                            flag = "DOC_BOXES";
                        }
                        // Time to stop grabbing document boxes?
                        else if (Regex.IsMatch(line, @"^#\*+calculations\*+$"))
                        {
                            flag = null;
                        }

                        //
                        // Looking for 'ETX' to indicate we are ready to start grabbing code boxes.
                        // The '$' and '\' characters are special and need to be prefaced with a '\'.
                        // The double quote characters inside the string need to be represented with
                        // 2 double quote chars for each double quote.
                        //
                        else if (Regex.IsMatch(line, @"\$\\=""\\x03\\n"""))
                        {
                            flag = "CODE_BOXES";
                        }
                        else
                        {
                            switch (flag)
                            {
                                case ("DOC_BOXES"):
                                    foreach (Match match in Regex.Matches(line, @"^(box\d+_\d+)"))
                                        doc_data.doc_box_list.Add(match.Value);
                                    break;

                                case ("CODE_BOXES"):

                                    // We need to check for an end of dataflow comment here since we
                                    // are blanking comment lines below.
                                    if (!end_of_df)
                                        end_of_df = Regex.IsMatch(line, @"^#[~#]*(((stop|end(of)?)dataflow)|(dataflow(end|stop)))");

                                    if (!profile_inst)
                                        profile_inst = Regex.IsMatch(line, @"^((individual=individual\.new)|(entity=entity\.new)|(preparer=preparer\.new)|(fiduciary=fiduciary\.new)|(equifax=equifax\.new))$");

                                    if (no_comments)
                                    {
                                        // Comment processing:
                                        //
                                        // We do not want to take into account any box references that are
                                        // inside a comment line. 99.9999% of calc comments start at index 0 of
                                        // a line, so safer to check index 0.
                                        if (line.IndexOf("#") == 0)
                                        {
                                            line = "";
                                        }
                                    }

                                    /*
                                     * eval() usage - Inside the new Check Box Control method is ok,
                                     * so we check first. Trying to think of everything here, there is
                                     * an unlikely possibility that we have 2 or more CheckBoxCtrl()
                                     * methods in the Ruby script file, so we run thru this list and
                                     * check for more than one instance.
                                    */
                                    if (Regex.IsMatch(line, @"eval\(.+\)"))
                                    {
                                        int firstLine, lastLine;
                                        string[] methodLines;
                                        bool inChkBoxRange = false;

                                        foreach (var method in methList)
                                        {
                                            /*
                                             * Each string is of the form: methodName:firstLine lastLine(char)254methodBody.
                                             * (char)254 is the separator between 'name and line info' and the method body.
                                            */
                                            methParts = method.Split(':');

                                            if (methParts[0] == "CheckBoxCtrl")
                                            {
                                                methodLines = methParts[1].Split((char)254)[0].Split(' ');

                                                if (int.TryParse(methodLines[0], out firstLine))
                                                {
                                                    if (int.TryParse(methodLines[1], out lastLine))
                                                    {
                                                        if (ln_num >= firstLine && ln_num <= lastLine)
                                                        {
                                                            inChkBoxRange = true;
                                                            break;
                                                        }
                                                    }
                                                }
                                            }
                                        }

                                        if (!inChkBoxRange)
                                        {
                                            // Add this to our doc_data.evalList to alert the user.
                                            doc_data.evalList.Add(ln_num);
                                        }
                                    }

                                    /*
                                     * Checklist Control Verification:
                                     * 
                                     * A regex match for both @"^chk_list=\[\[" and
                                     * @"^box\d+_\d+=state_list\.join\("",""\)" is our flag that we have
                                     * old check box control code in the Ruby code.
                                     * 
                                    */ 
                                    if (!chk_list)
                                    {
                                        chk_list = Regex.IsMatch(line, @"^chk_list=\[\[");
                                    }
                                    else if (!chk_list_join)
                                    {
                                        if (Regex.IsMatch(line, @"^box\d+_\d+=state_list\.join\("",""\)"))
                                        {
                                            // Set flag and grab state box.
                                            chk_list_join = true;
                                        }
                                    }

                                    if (Regex.IsMatch(line, @"box\d+_\d+"))
                                    {
                                        // print line processing:
                                        //
                                        // Check if this is a print line, if it is we will look for an
                                        // unbalanced print, i.e. "print 'box12_83', box12_82".
                                        if (Regex.IsMatch(line, @"^print'box\d+_\d+',box\d+_\d+$"))
                                        {
                                            MatchCollection boxes = Regex.Matches(line, @"box\d+_\d+");

                                            if (boxes[0].ToString() != boxes[1].ToString())
                                            {
                                                doc_data.unbal_print.Add(String.Format("{0}\nLine Number:{1}", line, ln_num));
                                            }
                                        }

                                        // Ruby code box references:
                                        //
                                        // Decided to keep track of ALL code boxes and also keep track
                                        // of all line numbers where each code box is used. Another routine
                                        // will determine which code boxes are not actually on the document
                                        // [if any] and will grab the line number(s).
                                        foreach (Match match in Regex.Matches(line, @"(box\d+_\d+)"))
                                        {

                                            // Insert code box if it is not in the code box list.
                                            if (!doc_data.code_box_list.Contains(match.Value))
                                                doc_data.code_box_list.Add(match.Value);

                                            // Add the box to code box dictionary if not already in there.
                                            if (doc_data.code_box_lines.ContainsKey(match.Value))
                                            {
                                                // We do not want duplicate line numbers for the box:
                                                if (!doc_data.code_box_lines[match.Value].Contains(ln_num))
                                                {
                                                    line_list = doc_data.code_box_lines[match.Value];
                                                    line_list.Add(ln_num);
                                                    doc_data.code_box_lines[match.Value] = line_list;
                                                }
                                            }
                                            else
                                            {
                                                List<int> ln_list = new List<int>();
                                                ln_list.Add(ln_num);
                                                doc_data.code_box_lines.Add(match.Value, ln_list);
                                            }
                                        }
                                    }
                                    break;
                            }
                        }
                    }

                    // Check for a missing end of dataflow comment with a profile instantiation.
                    if (profile_inst && !end_of_df)
                        doc_data.no_eodf_comment = true;

                    // Checklist Control: Checking for presence of older code.
                    doc_data.chk_box_ctrl_flag = (chk_list && chk_list_join);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Error reading file.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return doc_data;
        }

        // This is called from form1.cs [Form1()] when command line args have been supplied.
        // Also called from form1.cs [openToolStripMenuItem_Click()] after user selects a file
        // to inspect and after we run GetData().
        public void ProcessData(Data doc_info)
        {
            bool no_err = true;
            string err_msg;

            outText.Clear();

            // Checking if there are any boxes referenced in the Ruby Code that are not on the 
            // document:
            List<string> diff_list = doc_info.code_box_list.Except(doc_info.doc_box_list).ToList();

            if (diff_list.Count > 0)
            {
                no_err = false;
                err_msg = "ERROR! The Ruby Code references these boxes which are not on the document:\n";
                DisplayErrorMsg(err_msg, "ERR");

                // Each box in diff_list is a key to the Dictionary that holds each box's line number
                // (s).
                foreach (string box in diff_list)
                {
                    outText.AppendText(string.Format("\n{0} - Line Numbers: ", box));

                    // Being extra cautious here, this box should already be in the dictionary.
                    if (doc_info.code_box_lines.ContainsKey(box))
                    {
                        outText.AppendText(string.Join(", ", doc_info.code_box_lines[box].ToArray()));
                    }
                }
                outText.AppendText("\n\n");
            }

            // Checking if there are any unbalanced print statements:
            if (doc_info.unbal_print.Count > 0)
            {
                no_err = false;
                err_msg = "ERROR! The following print statements have non-matching boxes:\n\n";
                DisplayErrorMsg(err_msg, "ERR");

                foreach (string prnt_line in doc_info.unbal_print)
                {
                    outText.AppendText(string.Format("{0}\n\n", prnt_line));
                }
            }

            // Checking if an end of dataflow comment is missing. Treated as a warning.
            if (doc_info.no_eodf_comment)
            {
                no_err = false;
                err_msg = "WARNING! A profile has been instantiated but there is no end of dataflow comment.\n\n";
                DisplayErrorMsg(err_msg, "WAR");
            }

            // Checklist Control: Update old code?
            if (doc_info.chk_box_ctrl_flag)
            {
                no_err = false;
                err_msg = "ERROR! The Check Box Control code is an old version. Please replace with the current version if Check Box Control is still required, else delete the code.\n\n";
                DisplayErrorMsg(err_msg, "ERR");
            }

            // [PROPOSED] Checking for methods that need to be updated to Method Master File Version:
            //if (doc_info.methUpdate.Count > 0)
            //{
            //    no_err = false;
            //    err_msg = string.Format("ERROR! The following methods need to be UPDATED to the master version as specified in the\nMethod Master File: {0}\n", methMasterFile);
            //    DisplayErrorMsg(err_msg, "ERR");

            //    outText.AppendText("\n\tThe 'Update To' master name will be found in the specified category.\n\tPlease be aware that you may also need to alter the method's invocation code.\n");

            //    foreach (string str in doc_info.methUpdate)
            //    {
            //        outText.AppendText(string.Format("\n{0}", str));
            //    }

            //    outText.AppendText("\n\n");
            //}

            // [PROPOSED] Checking for methods that need to be added to the Method Master File:
            //if (doc_info.methAddMstr.Count > 0)
            //{
                //no_err = false;
                //err_msg = string.Format("ERROR! The following methods need to be ADDED to the Method Master File:\n{0}\n", methMasterFile);
                //DisplayErrorMsg(err_msg, "ERR");

                //// This will most likely need tweaking:
                //outText.AppendText("\n\tThe following steps will need to be followed:\n\n\t1) Notify the person in charge of adding new methods to the master file.\n\t2) Place the document on HOLD if instructed to do so.\n\t3) Update the method to the master file version if instructed to do so.\n");

                //foreach (string str in doc_info.methAddMstr)
                //{
                //    outText.AppendText(string.Format("\n{0}", str));
                //}

                //outText.AppendText("\n\n");
            //}

            // Checking for methods in the document with the same name:
            if (doc_info.dupMethName)
            {
                no_err = false;
                err_msg = "WARNING! Duplicate method names have been detected:\n";
                DisplayErrorMsg(err_msg, "WAR");

                foreach (KeyValuePair<string, List<string>> kvp in doc_info.methNames)
                {
                    if (kvp.Value.Count > 1)
                    {
                        outText.AppendText(string.Format("\n{0} - 'def' Line Numbers: {1}", kvp.Key, string.Join(", ", kvp.Value)));
                    }
                }

                outText.AppendText("\n\n");
            }

            // Checking for name_block( ) usage:
            if (doc_info.methNames.ContainsKey("name_block"))
            {
                no_err = false;
                err_msg = "WARNING! The name_block() method has been detected:\n";
                DisplayErrorMsg(err_msg, "WAR");

                foreach (KeyValuePair<string, List<string>> kvp in doc_info.methNames)
                {
                    if (kvp.Key == "name_block")
                    {
                        outText.AppendText(string.Format("\n{0} - 'def' Line Number: {1}\n", kvp.Key, string.Join(", ", kvp.Value)));
                    }
                }

                outText.AppendText("\n");
            }

            // Checking for eval( ) usage:
            if (doc_info.evalList.Count > 0)
            {
                no_err = false;
                err_msg = "WARNING! eval() usage has been detected on the following line(s) - Run the Black Box Converter from\nSuperBoxing to check for conversion errors:\n";

                DisplayErrorMsg(err_msg, "WAR");

                outText.AppendText(string.Format("\n{0}\n\n", string.Join(", ", doc_info.evalList)));
            }

            // Give the 'all clear' message if no errors detected.
            if (no_err)
            {
                outText.AppendText("Code Inspector Has Not Detected Any Issues.");
            }
        }

        public void ProcessMethods(ref Data doc_data)
        {
            /*
             * The Category arrays represent the method categorys we are inspecting 
             * to make sure they are up to date. The check digit routine has morphed over
             * the years into several versions, the amap routines were re-written around
             * 2011.
            */ 
            string[] cdCategory = { "check_digit", @"check_digit_\d+" };

            // These have historically been kept together under the 'amap' name.
            string[] amapCategory = { 
            @"amap_\d+", "delaware_amap", "delaware_cd", "fed_amap",
            "fed5500_cd", "utah_amap", "amap_suffix"
            };

            //string line;
            //string cat = "";
            string[] methParts;
            //string methName;
            //string mstrName;
            //string methMsg;
            //string methLine;
            //string methBody;
            //string md5HashStr;
            //bool checkMasterFile = false;
            //bool checkCat = false;
            //bool handled = false;
            //bool obChkSumMatch = false;
            //bool mstrChkSumMatch = false;
            //Match match;

            try
            {
                /*
                 * Get the method names from the Ruby script.
                 * 
                 * Each string in methList is structured like this
                 * example:
                 * 
                 * "name:10 20(char)254body"
                 * 
                 * Where the portion before the colon is the name, the
                 * 10 would be the line number of the 'def name...' 
                 * portion and a separating space char followed by the
                 * last line number, where the 'end' is, line 20 in this
                 * example. The (char)254 separates name and line number info
                 * from the actual method body, with whitespace and
                 * comment lines removed - I tried to pick a char that would not
                 * be in a method.
                 * 
                 * We use the ending line number when informing the user
                 * about eval( ) usage [in GetData( )] as we do not want
                 * to mention the presence of eval( ) in the Check Box Control
                 * method where it is is allowed.
                */
                methList = RubyExpressions.GetScriptExpressions(fname);

                //foreach (var method in methList)
                //{
                //    methParts = method.Split(':');
                //    methName = methParts[0];

                //    /*
                //     * There are many methods in our old ASCII Amap file,
                //     * I thought it was safer to list them all by name,
                //     * as many have the word 'amap' in their names, and some
                //     * do not even read like an amap method, but historically these 
                //     * have been bundled under the 'Amap' moniker.
                //    */ 
                //    foreach (var amapCat in amapCategory)
                //    {
                //        if (Regex.IsMatch(methName, amapCat))
                //        {
                //            cat = "amap";
                //            checkMasterFile = true;
                //            break;
                //        }
                //    }

                //    if (!checkMasterFile)
                //    {
                //        foreach (var cdCat in cdCategory)
                //        {
                //            if (Regex.IsMatch(methName, cdCat))
                //            {
                //                cat = "check_digit";
                //                checkMasterFile = true;
                //                break;
                //            }
                //        }
                //    }

                //    if (checkMasterFile)
                //    {
                //        methLine = methParts[1].Split(' ')[0];
                //        methBody = method.Split((char)254)[1];

                //        md5HashStr = md5(Encoding.ASCII.GetBytes(methBody)).ToString();

                //        //MessageBox.Show(string.Format("Method Name: {0}\nMethod Body:\n\"{1}\"\nChecksum: {2}", methName, methBody, md5HashStr));

                //        using (StreamReader sr = new StreamReader(methMasterFile))
                //        {
                //            while ((line = sr.ReadLine()) != null)
                //            {
                //                if (line == string.Format("<{0}>", cat))
                //                {
                //                    // The start of our category. Meow.
                //                    checkCat = true;
                //                }
                //                else if (line == string.Format("</{0}/>", cat))
                //                {
                //                    // The end of our category.
                //                    break;
                //                }
                //                else if (checkCat)
                //                {
                //                    // 'Obsolete' might not be the greatest name for this thing...
                //                    if (Regex.IsMatch(line, @"^Obsolete Checksum:"))
                //                    {
                //                        foreach (Match chkMatch in Regex.Matches(line, @"(\d+)"))
                //                        {
                //                            if (chkMatch.Value == md5HashStr)
                //                            {
                //                                /*
                //                                 * We now know we have matched the calculated checksum against
                //                                 * the 'Obsolete Checksum' in the proper category of our master file.
                //                                */
                //                                obChkSumMatch = true;
                //                                break;
                //                            }
                //                        }
                //                    }
                //                    else
                //                    {
                //                        match = Regex.Match(line, @"Master Checksum:""(\d+)""");

                //                        if (match.Success)
                //                        {
                //                            if (match.Groups[1].Value == md5HashStr)
                //                            {
                //                                /*
                //                                 * We now know we have matched the calculated checksum against
                //                                 * the 'Master Checksum' in the proper category of our master file.
                //                                */
                //                                mstrChkSumMatch = true;
                //                            }
                //                        }
                //                        else if (obChkSumMatch || mstrChkSumMatch)
                //                        {
                //                            // Now we can tell if we found the method or we need to update the method.
                //                            match = Regex.Match(line, @"Master Name:""(\w+)""");

                //                            if (match.Success)
                //                            {
                //                                //mstrName = match.Value;
                //                                mstrName = match.Groups[1].Value;

                //                                /*
                //                                 * If the 'Obsolete Checksum' passed: If the 'Master Checksum' passes we
                //                                 * know the document's method is fine. If the 'Master Checksum' fails we
                //                                 * know we have an older method version that needs updating.
                //                                 * 
                //                                 * If the 'Obsolete Checksum' failed: If the 'Master Checksum' passes we
                //                                 * know the document's method is fine, it is the current version. If the
                //                                 * 'Master Checksum' fails we know we need to keep looking for our method's
                //                                 * section in the master file, if it is in there at all.
                //                                */
                //                                if (obChkSumMatch)
                //                                {
                //                                    handled = true;

                //                                    if (!mstrChkSumMatch)
                //                                    {
                //                                        // Handled, but method needs an update.
                //                                        methMsg = string.Format("Method: {0}, 'def' Line: {1}, Update To: {2}, Category: {3} METHODS", methName, methLine, mstrName, cat.ToUpper());
                //                                        doc_data.methUpdate.Add(methMsg);
                //                                    }
                //                                }
                //                                else if (mstrChkSumMatch)
                //                                {
                //                                    // Method is current.
                //                                    handled = true;
                //                                }
                //                            }
                //                        }
                //                    }
                //                }

                //                if (handled)
                //                {
                //                    break;
                //                }
                //            }
                //        }

                //        if (!handled)
                //        {
                //            // Provide DocId, method name, line number and obChkSumMatch in error message.
                //            methMsg = string.Format("Method: {0}, 'def' Line: {1}, Category: {2} METHODS", methName, methLine, cat.ToUpper());
                //            doc_data.methAddMstr.Add(methMsg);
                //        }

                //        handled = false;
                //        mstrChkSumMatch = false;
                //        obChkSumMatch = false;
                //        checkCat = false;
                //        checkMasterFile = false;
                //    }
                //}

                // Now look for methods with the same name in the same document:
                List<string> methLines;

                foreach (var method in methList)
                {
                    methParts = method.Split(':');

                    if (doc_data.methNames.TryGetValue(methParts[0], out methLines) == true)
                    {
                        doc_data.dupMethName = true;

                        // Store only the starting line number.
                        methLines.Add(methParts[1].Split(' ')[0]);
                    }
                    else
                    {
                        methLines = new List<string>();
                        methLines.Add(methParts[1].Split(' ')[0]);
                        doc_data.methNames.Add(methParts[0], methLines);
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessMethods( ) Error...", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }
    }
}
